@@ -1,5 +1,5 @@
 /***************************************************************************
- *   Copyright (C) 2005 to 2010 by Jonathan Duddington                     *
+ *   Copyright (C) 2005 to 2011 by Jonathan Duddington                     *
  *   email: jonsd@users.sourceforge.net                                    *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
@@ -39,24 +39,35 @@
 #define L_qa   0x716100
 #define L_grc  0x677263   // grc  Ancient Greek
 #define L_jbo  0x6a626f   // jbo  Lojban
-#define L_pap  0x706170   // pap  Papiamento
+#define L_pap  0x706170   // pap  Papiamento]
+#define L_qvi  0x717669   // qvi  Kichwa
+#define L_shs  0x736873   // shs  Shuswap / Secwepemctsin
 #define L_zhy  0x7a6879   // zhy
 
 // start of unicode pages for character sets
-#define OFFSET_GREEK  0x380
+#define OFFSET_GREEK    0x380
 #define OFFSET_CYRILLIC 0x420
 #define OFFSET_ARMENIAN 0x530
-#define OFFSET_ARABIC 0x600
+#define OFFSET_ARABIC   0x600
+#define OFFSET_THAANA   0x780      // Divehi/Maldives
 #define OFFSET_DEVANAGARI  0x900
-#define OFFSET_BENGALI 0x980
+#define OFFSET_BENGALI  0x980
 #define OFFSET_GURMUKHI 0xa00
-#define OFFSET_TAMIL  0xb80
-#define OFFSET_KANNADA 0xc80
+#define OFFSET_GUJARATI 0xa80
+#define OFFSET_ORIYA    0xb00
+#define OFFSET_TAMIL    0xb80
+#define OFFSET_TELUGU   0xc00
+#define OFFSET_KANNADA  0xc80
 #define OFFSET_MALAYALAM 0xd00
-#define OFFSET_KOREAN 0x1100
+#define OFFSET_SINHALA  0x0d80
+#define OFFSET_GEORGIAN 0x1080
+
+#define OFFSET_KOREAN   0x1100
 
 static void Translator_Russian(Translator *tr);
 
+
+
 static void SetLetterVowel(Translator *tr, int c)
 {//==============================================
 	tr->letter_bits[c] = (tr->letter_bits[c] & 0x40) | 0x81;  // keep value for group 6 (front vowels e,i,y)
@@ -107,6 +118,7 @@ static Translator* NewTranslator(void)
 	static const short stress_lengths2[8] = {182,140, 220,220, 220,240, 260,280};
 	static const wchar_t empty_wstring[1] = {0};
 	static const wchar_t punct_in_word[2] = {'\'', 0};  // allow hyphen within words
+	static const unsigned char default_tunes[6] = {0, 1, 2, 3, 0, 0};
 
 	tr = (Translator *)Alloc(sizeof(Translator));
 	if(tr == NULL)
@@ -118,14 +130,16 @@ static Translator* NewTranslator(void)
 	tr->data_dictrules = NULL;     // language_1   translation rules file
 	tr->data_dictlist = NULL;      // language_2   dictionary lookup file
 
-	tr->transpose_offset = 0;
+	tr->transpose_min = 'a';
+	tr->transpose_max = 'z';
+	tr->frequent_pairs = NULL;
 
 	// only need lower case
 	tr->letter_bits_offset = 0;
 	memset(tr->letter_bits,0,sizeof(tr->letter_bits));
 	memset(tr->letter_groups,0,sizeof(tr->letter_groups));
 
-	// 0-5 sets of characters matched by A B C E F G in pronunciation rules
+	// 0-5 sets of characters matched by A B C H F G Y  in pronunciation rules
 	// these may be set differently for different languages
 	SetLetterBits(tr,0,"aeiou");  // A  vowels, except y
 	SetLetterBits(tr,1,"bcdfgjklmnpqstvxz");      // B  hard consonants, excluding h,r,w
@@ -153,15 +167,16 @@ static Translator* NewTranslator(void)
 	tr->langopts.unstressed_wd1 = 1;
 	tr->langopts.unstressed_wd2 = 3;
 	tr->langopts.param[LOPT_SONORANT_MIN] = 95;
+	tr->langopts.param[LOPT_LONG_VOWEL_THRESHOLD] = 190/2;
 	tr->langopts.param[LOPT_MAXAMP_EOC] = 19;
 	tr->langopts.param[LOPT_UNPRONOUNCABLE] = 's';    // don't count this character at start of word
 	tr->langopts.param[LOPT_BRACKET_PAUSE] = 4;      // pause at bracket
 	tr->langopts.param2[LOPT_BRACKET_PAUSE] = 2;    // pauses when announcing bracket names
-	tr->langopts.param[LOPT_MIN_LONG_VOWEL] = 40;
 	tr->langopts.max_initial_consonants = 3;
 	tr->langopts.replace_chars = NULL;
 	tr->langopts.ascii_language = "";    // Non-Latin alphabet languages, use this language to speak Latin words, default is English
 
+
 	SetLengthMods(tr,201);
 //	tr->langopts.length_mods = length_mods_en;
 //	tr->langopts.length_mods0 = length_mods_en0;
@@ -173,12 +188,53 @@ static Translator* NewTranslator(void)
 	tr->langopts.thousands_sep = ',';
 	tr->langopts.decimal_sep = '.';
 	tr->langopts.break_numbers = BREAK_THOUSANDS;   // 1000, 1000,000  1,000,000 etc
+	tr->langopts.max_digits = 14;
 
 	memcpy(tr->punct_to_tone, punctuation_to_tone, sizeof(tr->punct_to_tone));
 
+	memcpy(tr->langopts.tunes, default_tunes, sizeof(tr->langopts.tunes));
+
 	return(tr);
 }
 
+// common letter pairs, encode these as a single byte
+//  2 bytes, using the transposed character codes
+static const short pairs_ru[] = { 
+0x010c, //  ла   21052  0x23
+0x010e, //  на   18400
+0x0113, //  та   14254
+0x0301, //  ав   31083
+0x030f, //  ов   13420
+0x060e, //  не   21798
+0x0611, //  ре   19458
+0x0903, //  ви   16226
+0x0b01, //  ак   14456
+0x0b0f, //  ок   17836
+0x0c01, //  ал   13324
+0x0c09, //  ил   16877
+0x0e01, //  ан   15359
+0x0e06, //  ен   13543  0x30
+0x0e09, //  ин   17168
+0x0e0e, //  нн   15973
+0x0e0f, //  он   22373
+0x0e1c, //  ын   15052
+0x0f03, //  во   24947
+0x0f11, //  ро   13552
+0x0f12, //  со   16368
+0x100f, //  оп   19054
+0x1011, //  рп   17067
+0x1101, //  ар   23967
+0x1106, //  ер   18795
+0x1109, //  ир   13797
+0x110f, //  ор   21737
+0x1213, //  тс   25076
+0x1220, //  яс   14310
+0x7fff};
+//0x040f  ог   12976
+//0x1306  ет   12826
+//0x0f0d  мо   12688
+
+
 
 static const unsigned int replace_cyrillic_latin[] = 
 	{0x430,'a',
@@ -216,6 +272,33 @@ static const unsigned int replace_cyrillic_latin[] =
 	0x45c,0x107,
 0};  // ѓ  ѕ  ќ
 
+static void SetCyrillicLetters(Translator *tr)
+{//===========================================
+	// character codes offset by 0x420
+	static const char ru_vowels[] = {0x10,0x15,0x31,0x18,0x1e,0x23,0x2b,0x2d,0x2e,0x2f,0};
+	static const char ru_consonants[] = {0x11,0x12,0x13,0x14,0x16,0x17,0x19,0x1a,0x1b,0x1c,0x1d,0x1f,0x20,0x21,0x22,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2c,0};
+	static const char ru_soft[] = {0x2c,0x19,0x27,0x29,0};   // letter group B  [k ts; s;]
+	static const char ru_hard[] = {0x2a,0x16,0x26,0x28,0};   // letter group H  [S Z ts]
+	static const char ru_nothard[] = {0x11,0x12,0x13,0x14,0x17,0x19,0x1a,0x1b,0x1c,0x1d,0x1f,0x20,0x21,0x22,0x24,0x25,0x27,0x29,0x2c,0};
+	static const char ru_voiced[] = {0x11,0x12,0x13,0x14,0x16,0x17,0};    // letter group G  (voiced obstruents)
+	static const char ru_ivowels[] = {0x2c,0x2e,0x2f,0x31,0};   // letter group Y  (iotated vowels & soft-sign)
+	tr->charset_a0 = charsets[18];   // KOI8-R
+	tr->transpose_min = 0x430;  // convert cyrillic from unicode into range 0x01 to 0x22
+	tr->transpose_max = 0x451;
+	tr->frequent_pairs = pairs_ru;
+
+	tr->letter_bits_offset = OFFSET_CYRILLIC;
+	memset(tr->letter_bits,0,sizeof(tr->letter_bits));
+	SetLetterBits(tr,LETTERGP_A,ru_vowels);
+	SetLetterBits(tr,1,ru_soft);
+	SetLetterBits(tr,2,ru_consonants);
+	SetLetterBits(tr,3,ru_hard);
+	SetLetterBits(tr,4,ru_nothard);
+	SetLetterBits(tr,5,ru_voiced);
+	SetLetterBits(tr,6,ru_ivowels);
+	SetLetterBits(tr,LETTERGP_VOWEL2,ru_vowels);
+}  // end of SetCyrillicLetters
+
 
 void SetIndicLetters(Translator *tr)
 {//=================================
@@ -238,6 +321,7 @@ void SetIndicLetters(Translator *tr)
 	tr->langopts.suffix_add_e = tr->letter_bits_offset + 0x4d;   //virama
 }
 
+
 void SetupTranslator(Translator *tr, const short *lengths, const unsigned char *amps)
 {//==================================================================================
 	if(lengths != NULL)
@@ -252,9 +336,15 @@ Translator *SelectTranslator(const char *name)
 	int name2 = 0;
 	Translator *tr;
 
+	static const short stress_lengths_fr[8] = {190, 170,  190, 200,  0, 0,  190, 240};
+	static const unsigned char stress_amps_fr[8] = {18,16, 18,18, 18,18, 18,18 };
+
 	static const unsigned char stress_amps_sk[8] = {17,17, 20,20, 20,22, 22,21 };
 	static const short stress_lengths_sk[8] = {190,190, 210,210, 0,0, 210,210};
 
+	static const short stress_lengths_ta[8] = {200, 200,  210, 210,  0, 0,  230, 230};
+	static const unsigned char stress_amps_ta[8] = {18,18, 18,18, 20,20, 22,22 };
+
 	// convert name string into a word of up to 4 characters, for the switch()
 	while(*name != 0)
 		name2 = (name2 << 8) + *name++;
@@ -279,11 +369,35 @@ Translator *SelectTranslator(const char *name)
 		}
 		break;
 
+	case L('a','m'):  // Amharic, Ethiopia
+		{
+			SetupTranslator(tr,stress_lengths_fr,stress_amps_fr);
+			tr->langopts.stress_rule = STRESSPOSN_1L;
+			tr->langopts.stress_flags = 0x0024;  // don't use secondary stress
+			tr->langopts.length_mods0 = tr->langopts.length_mods;  // don't lengthen vowels in the last syllable
+			tr->langopts.param[LOPT_UNPRONOUNCABLE] = 1;   // disable check for unpronouncable words
+		}
+		break;
+
+
 	case L('a','r'):   // Arabic
 			tr->letter_bits_offset = OFFSET_ARABIC;
 			tr->langopts.param[LOPT_UNPRONOUNCABLE] = 1;   // disable check for unpronouncable words
 		break;
 
+	case L('b','g'):  //Bulgarian
+		{
+			SetCyrillicLetters(tr);
+			SetLetterVowel(tr,0x2a);
+			tr->langopts.param[LOPT_UNPRONOUNCABLE] = 0x432;    // [v]  don't count this character at start of word
+			tr->langopts.param[LOPT_REGRESSIVE_VOICING] = 0x10;  // devoice at end of word
+			tr->langopts.param[LOPT_REDUCE] = 2;
+			tr->langopts.stress_rule = STRESSPOSN_2R;
+			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_ALLOW_SPACE | NUM_OMIT_1_HUNDRED | NUM_HUNDRED_AND | NUM_AND_UNITS | NUM_SINGLE_AND | NUM_ROMAN | NUM_ROMAN_ORDINAL | NUM_ROMAN_CAPITALS ;
+			tr->langopts.thousands_sep = ' ';   // don't allow dot as thousands separator
+		}
+		break;
+
 	case L('b','n'):  // Bengali
 		{
 			static const short stress_lengths_bn[8] = {180, 180,  210, 210,  0, 0,  230, 240};
@@ -330,27 +444,45 @@ Translator *SelectTranslator(const char *name)
 
 	case L('d','a'):  // Danish
 		{
-			static const short stress_lengths_da[8] = {160,140, 200,200, 0,0, 220,210};
+			static const short stress_lengths_da[8] = {160,140, 200,200, 0,0, 220,230};
 			SetupTranslator(tr,stress_lengths_da,NULL);
 
 			tr->langopts.stress_rule = STRESSPOSN_1L;
+			tr->langopts.param[LOPT_PREFIXES] = 1;
 			SetLetterVowel(tr,'y');
-			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_SWAP_TENS | NUM_HUNDRED_AND | NUM_OMIT_1_HUNDRED | NUM_ORDINAL_DOT | NUM_1900;
+			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_SWAP_TENS | NUM_HUNDRED_AND | NUM_OMIT_1_HUNDRED | NUM_ORDINAL_DOT | NUM_1900 | NUM_ROMAN | NUM_ROMAN_CAPITALS | NUM_ROMAN_ORDINAL;
 		}
 		break;
 
 
 	case L('d','e'):
 		{
-			static const short stress_lengths_de[8] = {150,130, 200,200,  0, 0, 260,275};
+			static const short stress_lengths_de[8] = {150,130, 200,200,  0, 0, 250,260};
 			tr->langopts.stress_rule = STRESSPOSN_1L;
 			tr->langopts.word_gap = 0x8;   // don't use linking phonemes
 			tr->langopts.vowel_pause = 0x30;
 			tr->langopts.param[LOPT_PREFIXES] = 1;
+			tr->langopts.param[LOPT_REGRESSIVE_VOICING] = 0x10;  // devoice at end of word
+			tr->langopts.param[LOPT_LONG_VOWEL_THRESHOLD] = 175/2;
 			memcpy(tr->stress_lengths,stress_lengths_de,sizeof(tr->stress_lengths));
 		
-			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_SWAP_TENS | NUM_OMIT_1_HUNDRED | NUM_ALLOW_SPACE | NUM_ORDINAL_DOT | NUM_ROMAN;
+			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_SWAP_TENS | NUM_OMIT_1_HUNDRED | NUM_OMIT_1_THOUSAND | NUM_ALLOW_SPACE | NUM_ORDINAL_DOT | NUM_ROMAN;
 			SetLetterVowel(tr,'y');
+			tr->langopts.param[LOPT_UNPRONOUNCABLE] = 2;   // use de_rules for unpronouncable rules
+		}
+		break;
+
+	case L('d','v'):   // Divehi (Maldives)
+		{
+			SetupTranslator(tr,stress_lengths_ta,stress_amps_ta);
+			tr->langopts.param[LOPT_UNPRONOUNCABLE] = 1;   // disable check for unpronouncable words
+			tr->langopts.length_mods0 = tr->langopts.length_mods;  // don't lengthen vowels in the last syllable
+			tr->letter_bits_offset = OFFSET_THAANA;
+			tr->langopts.stress_rule = STRESSPOSN_1L;
+			tr->langopts.stress_flags =  0x10004;   // use 'diminished' for unstressed final syllable
+			SetLetterBitsRange(tr,LETTERGP_B,0x26,0x30);   // vowel signs, and virama
+			tr->langopts.break_numbers = 0x14a8;  // 1000, 100,000  10,000,000 
+			tr->langopts.numbers = 1;
 		}
 		break;
 
@@ -360,9 +492,12 @@ Translator *SelectTranslator(const char *name)
 			SetupTranslator(tr,stress_lengths_en,NULL);
 
 			tr->langopts.stress_rule = STRESSPOSN_1L;
+			tr->langopts.stress_flags = 0x08;
 			tr->langopts.numbers = NUM_HUNDRED_AND | NUM_ROMAN | NUM_1900;
 			tr->langopts.param[LOPT_COMBINE_WORDS] = 2;       // allow "mc" to cmbine with the following word
 			tr->langopts.suffix_add_e = 'e';
+			tr->langopts.param[LOPT_UNPRONOUNCABLE] = 2;   // use en_rules for unpronouncable rules
+			SetLetterBits(tr,6,"aeiouy");  // Group Y: vowels, including y
 		}
 		break;
 
@@ -387,6 +522,7 @@ Translator *SelectTranslator(const char *name)
 			tr->letter_bits_offset = OFFSET_GREEK;
 			memset(tr->letter_bits,0,sizeof(tr->letter_bits));
 			SetLetterBits(tr,LETTERGP_A,el_vowels);
+			SetLetterBits(tr,LETTERGP_VOWEL2,el_vowels);
 			SetLetterBits(tr,LETTERGP_B,el_voiceless);
 			SetLetterBits(tr,LETTERGP_C,el_consonants);
 			SetLetterBits(tr,LETTERGP_Y,el_fvowels);    // front vowels: ε η ι υ
@@ -399,7 +535,7 @@ Translator *SelectTranslator(const char *name)
 			tr->langopts.param[LOPT_SONORANT_MIN] = 130;  // limit the shortening of sonorants before short vowels
 
 			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_DECIMAL_COMMA; 
-			tr->langopts.numbers2 = 0x2;   // variant form of numbers before thousands
+			tr->langopts.numbers2 = 0x2 | NUM2_MULTIPLE_ORDINAL;   // variant form of numbers before thousands
 
 			if(name2 == L_grc)
 			{
@@ -411,7 +547,7 @@ Translator *SelectTranslator(const char *name)
 
 	case L('e','o'):
 		{
-			static const short stress_lengths_eo[8] = {145, 145,  230, 170,    0,   0,  360, 370};
+			static const short stress_lengths_eo[8] = {150, 150,  230, 180,    0,   0,  300, 320};
 			static const unsigned char stress_amps_eo[] = {16,14, 20,20, 20,22, 22,21 };
 			static const wchar_t eo_char_apostrophe[2] = {'l',0};
 		
@@ -420,11 +556,11 @@ Translator *SelectTranslator(const char *name)
 			tr->charset_a0 = charsets[3];  // ISO-8859-3
 			tr->char_plus_apostrophe = eo_char_apostrophe;
 
-			tr->langopts.word_gap = 1;
+//			tr->langopts.word_gap = 1;
 			tr->langopts.vowel_pause = 2;
 			tr->langopts.stress_rule = STRESSPOSN_2R;
 			tr->langopts.stress_flags =  0x6 | 0x10; 
-			tr->langopts.unstressed_wd1 = 3;
+//			tr->langopts.unstressed_wd1 = 3;
 			tr->langopts.unstressed_wd2 = 2;
 
 			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_OMIT_1_HUNDRED | NUM_ALLOW_SPACE | NUM_ROMAN;
@@ -452,7 +588,8 @@ Translator *SelectTranslator(const char *name)
 			tr->langopts.unstressed_wd2 = 2;
 			tr->langopts.param[LOPT_SONORANT_MIN] = 120;  // limit the shortening of sonorants before short vowels
 
-			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_DECIMAL_COMMA | NUM_AND_UNITS | NUM_OMIT_1_HUNDRED | NUM_ROMAN | NUM_ROMAN_AFTER;
+			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_DECIMAL_COMMA | NUM_AND_UNITS | NUM_OMIT_1_HUNDRED | NUM_OMIT_1_THOUSAND | NUM_ROMAN | NUM_ROMAN_AFTER;
+			tr->langopts.numbers2 = NUM2_MULTIPLE_ORDINAL; 
 
 			if(name2 == L('c','a'))
 			{
@@ -464,6 +601,10 @@ Translator *SelectTranslator(const char *name)
 			{
 				tr->langopts.stress_flags = 0x100 | 0x6 | 0x30;  // stress last syllable unless word ends with a vowel
 			}
+			else
+			{
+				tr->langopts.param[LOPT_UNPRONOUNCABLE] = 2;   // use es_rules for unpronouncable rules
+			}
 		}
 		break;
 
@@ -499,7 +640,6 @@ Translator *SelectTranslator(const char *name)
 			tr->langopts.stress_flags = 0x56;  // move secondary stress from light to a following heavy syllable
 			tr->langopts.param[LOPT_IT_DOUBLING] = 1;
 			tr->langopts.long_stop = 130;
-			tr->langopts.param[LOPT_MIN_LONG_VOWEL] = 65;
 
 			tr->langopts.numbers = NUM_DECIMAL_COMMA + NUM_ALLOW_SPACE;
 			SetLetterVowel(tr,'y');
@@ -509,18 +649,16 @@ Translator *SelectTranslator(const char *name)
 		}
 		break;
 
-
 	case L('f','r'):  // french
 		{
-			static const short stress_lengths_fr[8] = {190, 170,  190, 200,  0, 0,  235, 240};
-			static const unsigned char stress_amps_fr[8] = {18,16, 20,20, 20,22, 22,21 };
-
 			SetupTranslator(tr,stress_lengths_fr,stress_amps_fr);
 			tr->langopts.stress_rule = STRESSPOSN_1R;      // stress on final syllable
 			tr->langopts.stress_flags = 0x0024;  // don't use secondary stress
 			tr->langopts.param[LOPT_IT_LENGTHEN] = 1;    // remove lengthen indicator from unstressed syllables
+			tr->langopts.length_mods0 = tr->langopts.length_mods;  // don't lengthen vowels in the last syllable
+			tr->langopts.accents = 2;   // Say "Capital" after the letter.
 
-			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_DECIMAL_COMMA | NUM_ALLOW_SPACE | NUM_OMIT_1_HUNDRED | NUM_NOPAUSE | NUM_ROMAN | NUM_VIGESIMAL | NUM_DFRACTION_4;
+			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_DECIMAL_COMMA | NUM_ALLOW_SPACE | NUM_OMIT_1_HUNDRED | NUM_NOPAUSE | NUM_ROMAN | NUM_ROMAN_CAPITALS | NUM_ROMAN_AFTER | NUM_VIGESIMAL | NUM_DFRACTION_4;
 			SetLetterVowel(tr,'y');
 		}
 		break;
@@ -572,7 +710,7 @@ Translator *SelectTranslator(const char *name)
 			tr->langopts.spelling_stress = 1;
 			tr->langopts.accents = 1;
 
-			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_HUNDRED_AND | NUM_OMIT_1_HUNDRED | NUM_DECIMAL_COMMA | NUM_THOUS_SPACE | NUM_DFRACTION_2 | NUM_ROMAN_UC;
+			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_HUNDRED_AND | NUM_OMIT_1_HUNDRED | NUM_DECIMAL_COMMA | NUM_THOUS_SPACE | NUM_DFRACTION_2 | NUM_ROMAN_CAPITALS;
 			tr->langopts.numbers2 = 0x4a;  // variant numbers before thousands,milliards
 			tr->langopts.replace_chars = replace_cyrillic_latin;
 
@@ -582,6 +720,14 @@ Translator *SelectTranslator(const char *name)
 		break;
 
 
+	case L('h','t'):  // Haitian Creole
+//			memcpy(tr->stress_lengths,stress_lengths_fr,sizeof(tr->stress_lengths));
+			tr->langopts.stress_rule = STRESSPOSN_1R;      // stress on final syllable
+			tr->langopts.stress_flags = 0x0024;  // don't use secondary stress
+			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_OMIT_1_HUNDRED | NUM_NOPAUSE | NUM_ROMAN | NUM_VIGESIMAL | NUM_DFRACTION_4;
+		break;
+
+
 	case L('h','u'):   // Hungarian
 		{
 			static const unsigned char stress_amps_hu[8] = {17,17, 19,19, 20,22, 22,21 };
@@ -596,11 +742,11 @@ Translator *SelectTranslator(const char *name)
 			tr->langopts.unstressed_wd1 = 2;
 			tr->langopts.param[LOPT_IT_DOUBLING] = 1;
 			tr->langopts.param[LOPT_ANNOUNCE_PUNCT] = 2;  // don't break clause before announcing . ? !
-			tr->langopts.param[LOPT_MIN_LONG_VOWEL] = 65;
 
-			tr->langopts.numbers = NUM_DFRACTION_5 | NUM_ALLOW_SPACE | NUM_ROMAN | NUM_ROMAN_ORDINAL | NUM_ROMAN_CAPITALS | NUM_ORDINAL_DOT | NUM_OMIT_1_HUNDRED;
+			tr->langopts.numbers = NUM_DFRACTION_5 | NUM_ALLOW_SPACE | NUM_ROMAN | NUM_ROMAN_ORDINAL | NUM_ROMAN_CAPITALS | NUM_ORDINAL_DOT | NUM_OMIT_1_HUNDRED | NUM_OMIT_1_THOUSAND;
 			tr->langopts.thousands_sep = ' ';   // don't allow dot as thousands separator
 			tr->langopts.decimal_sep = ',';
+			tr->langopts.max_roman = 899;
 			tr->langopts.min_roman = 1;
 			SetLetterVowel(tr,'y');
 			tr->langopts.spelling_stress = 1;
@@ -621,6 +767,7 @@ SetLengthMods(tr,3);  // all equal
 			tr->letter_bits_offset = OFFSET_ARMENIAN;
 			memset(tr->letter_bits,0,sizeof(tr->letter_bits));
 			SetLetterBits(tr,LETTERGP_A,hy_vowels);
+			SetLetterBits(tr,LETTERGP_VOWEL2,hy_vowels);
 			SetLetterBits(tr,LETTERGP_C,hy_consonants);
 			tr->langopts.max_initial_consonants = 6;
 			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_OMIT_1_HUNDRED;
@@ -702,6 +849,23 @@ SetLengthMods(tr,3);  // all equal
 		}
 		break;
 
+	case L('k','a'):   // Georgian
+		{
+			// character codes offset by 0x1080
+			static const char ka_vowels[] = {0x50,0x54,0x58,0x5d,0x63,0x75,0x77,0};
+			SetupTranslator(tr,stress_lengths_ta,stress_amps_ta);
+			memset(tr->letter_bits,0,sizeof(tr->letter_bits));
+			SetLetterBits(tr,LETTERGP_A,ka_vowels);
+			SetLetterBits(tr,LETTERGP_VOWEL2,ka_vowels);
+
+			tr->langopts.stress_rule = STRESSPOSN_1L;
+			tr->letter_bits_offset = OFFSET_GEORGIAN;
+//			tr->langopts.param[LOPT_UNPRONOUNCABLE] = 1;   // disable check for unpronouncable words
+			tr->langopts.max_initial_consonants = 6;
+			tr->langopts.numbers = NUM_VIGESIMAL | NUM_AND_UNITS | NUM_OMIT_1_HUNDRED |NUM_OMIT_1_THOUSAND;
+		}
+		break;
+
 	case L('k','o'):   // Korean, TEST
 		{
 			static const char ko_ivowels[] = {0x63,0x64,0x67,0x68,0x6d,0x72,0x74,0x75,0};  // y and i vowels
@@ -771,10 +935,10 @@ SetLengthMods(tr,3);  // all equal
 
 			SetupTranslator(tr,stress_lengths_mk,stress_amps_mk);
 			tr->charset_a0 = charsets[5];   // ISO-8859-5
-			tr->letter_groups[0] = vowels_cyrillic;
+			tr->letter_groups[0] = tr->letter_groups[7] = vowels_cyrillic;
 
 			tr->langopts.stress_rule = STRESSPOSN_3R;   // antipenultimate
-			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_AND_UNITS | NUM_OMIT_1_HUNDRED | NUM_DFRACTION_2;
+			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_AND_UNITS | NUM_OMIT_1_HUNDRED | NUM_OMIT_1_THOUSAND | NUM_DFRACTION_2;
 			tr->langopts.numbers2 = 0x8a;  // variant numbers before thousands,milliards
 		}
 		break;
@@ -785,12 +949,15 @@ SetLengthMods(tr,3);  // all equal
 			static const short stress_lengths_nl[8] = {160,135, 210,210,  0, 0, 260,280};
 
 			tr->langopts.stress_rule = STRESSPOSN_1L;
-			tr->langopts.vowel_pause = 1;
+			tr->langopts.vowel_pause = 0x30;  // ??
 			tr->langopts.param[LOPT_DIERESES] = 1;
 			tr->langopts.param[LOPT_PREFIXES] = 1;
+			tr->langopts.param[LOPT_REGRESSIVE_VOICING] = 0x10;  // devoice at end of word
 			SetLetterVowel(tr,'y');
-		
-			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_SWAP_TENS | NUM_OMIT_1_HUNDRED | NUM_ALLOW_SPACE | NUM_1900 | NUM_ORDINAL_DOT;
+
+			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_SWAP_TENS | NUM_OMIT_1_HUNDRED | NUM_OMIT_1_THOUSAND | NUM_ALLOW_SPACE | NUM_1900 | NUM_ORDINAL_DOT;
+			tr->langopts.ordinal_indicator = "e";
+			tr->langopts.stress_flags = S_FIRST_PRIMARY;
 			memcpy(tr->stress_lengths,stress_lengths_nl,sizeof(tr->stress_lengths));
 		}
 		break;
@@ -917,6 +1084,17 @@ SetLengthMods(tr,3);  // all equal
 		}
 		break;
 
+	case L('s','l'):  // Slovenian
+			tr->charset_a0 = charsets[2];   // ISO-8859-2
+			tr->langopts.stress_rule = STRESSPOSN_2R;   // Temporary
+			tr->langopts.stress_flags = 0x20;
+			tr->langopts.param[LOPT_REGRESSIVE_VOICING] = 0x13;
+			tr->langopts.numbers =  NUM_DECIMAL_COMMA | NUM_ALLOW_SPACE | NUM_SWAP_TENS | NUM_OMIT_1_HUNDRED | NUM_DFRACTION_2 | NUM_ORDINAL_DOT | NUM_ROMAN;
+			tr->langopts.numbers2 = 0x100;   // plural forms of millions etc
+			tr->langopts.thousands_sep = ' ';   // don't allow dot as thousands separator
+			SetLetterVowel(tr,'r');
+		break;
+
 	case L('s','q'):  // Albanian
 		{
 			static const short stress_lengths_sq[8] = {150, 150,  180, 180,  0, 0,  300, 300};
@@ -947,6 +1125,7 @@ SetLengthMods(tr,3);  // all equal
 		break;
 
 	case L('s','w'):  // Swahili
+	case L('t','n'):  // Setswana
 		{
 			static const short stress_lengths_sw[8] = {160, 170,  200, 200,    0,   0,  320, 340};
 			static const unsigned char stress_amps_sw[] = {16,12, 19,19, 20,22, 22,21 };
@@ -967,21 +1146,20 @@ SetLengthMods(tr,3);  // all equal
 	case L('m','l'):  // Malayalam
 	case L('k','n'):  // Kannada
 	case L('m','r'):  // Marathi
+	case L('t','e'):  // Telugu
 		{
-			static const short stress_lengths_ta[8] = {200, 200,  210, 210,  0, 0,  230, 230};
-			static const unsigned char stress_amps_ta[8] = {18,18, 18,18, 20,20, 22,22 };
-
 			SetupTranslator(tr,stress_lengths_ta,stress_amps_ta);
 			tr->langopts.length_mods0 = tr->langopts.length_mods;  // don't lengthen vowels in the last syllable
 
 			tr->langopts.stress_rule = STRESSPOSN_1L;
 			tr->langopts.stress_flags =  0x10004;   // use 'diminished' for unstressed final syllable
+			tr->langopts.spelling_stress = 1;
 			tr->langopts.break_numbers = 0x14a8;  // 1000, 100,000  10,000,000 
 
 			if(name2 == L('t','a'))
 			{
 				tr->letter_bits_offset = OFFSET_TAMIL;
-				tr->langopts.numbers = 0x1;
+				tr->langopts.numbers =  NUM_OMIT_1_THOUSAND ;
 			}
 			if(name2 == L('m','r'))
 			{
@@ -991,6 +1169,7 @@ SetLengthMods(tr,3);  // all equal
 			if(name2 == L('m','l'))
 			{
 				tr->letter_bits_offset = OFFSET_MALAYALAM;
+				tr->langopts.numbers = NUM_OMIT_1_THOUSAND;
 			}
 			else
 			if(name2 == L('k','n'))
@@ -998,6 +1177,12 @@ SetLengthMods(tr,3);  // all equal
 				tr->letter_bits_offset = OFFSET_KANNADA;
 				tr->langopts.numbers = 0x1;
 			}
+			else
+			if(name2 == L('t','e'))
+			{
+				tr->letter_bits_offset = OFFSET_TELUGU;
+				tr->langopts.numbers = 0x1;
+			}
 			tr->langopts.param[LOPT_WORD_MERGE] = 1;   // don't break vowels betwen words
 			SetIndicLetters(tr);   // call this after setting OFFSET_
 		}
@@ -1023,8 +1208,8 @@ SetLengthMods(tr,3);  // all equal
 
 	case L('t','r'):   // Turkish
 		{
-			static const unsigned char stress_amps_tr[8] = {18,18, 20,20, 20,22, 22,21 };
-			static const short stress_lengths_tr[8] = {190,190, 190,190, 0,0, 250,270};
+			static const unsigned char stress_amps_tr[8] = {18,16, 20,21, 20,21, 21,20 };
+			static const short stress_lengths_tr[8] = {190,180, 200,230, 0,0, 240,250};
 
 			SetupTranslator(tr,stress_lengths_tr,stress_amps_tr);
 			tr->charset_a0 = charsets[9];   // ISO-8859-9 - Latin5
@@ -1032,7 +1217,7 @@ SetLengthMods(tr,3);  // all equal
 			tr->langopts.stress_rule = 7;   // stress on the last syllable, before any explicitly unstressed syllable
 			tr->langopts.stress_flags = 0x20;  //no automatic secondary stress
 
-			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_DECIMAL_COMMA | NUM_ALLOW_SPACE | NUM_OMIT_1_HUNDRED | NUM_DFRACTION_2;
+			tr->langopts.numbers = NUM_SINGLE_STRESS | NUM_DECIMAL_COMMA | NUM_ALLOW_SPACE | NUM_OMIT_1_HUNDRED | NUM_OMIT_1_THOUSAND | NUM_DFRACTION_2;
 			tr->langopts.max_initial_consonants = 2;
 		}
 		break;
@@ -1068,14 +1253,19 @@ SetLengthMods(tr,3);  // all equal
 			tr->langopts.stress_rule = STRESSPOSN_1L;
 			tr->langopts.word_gap = 0x21;   // length of a final vowel is less dependent on the next consonant, don't merge consonant with next word
 //			tr->langopts.vowel_pause = 4;
-			tr->letter_groups[0] = vowels_vi;
+			tr->letter_groups[0] = tr->letter_groups[7] = vowels_vi;
 			tr->langopts.tone_language = 1;   // Tone language, use  CalcPitches_Tone() rather than CalcPitches()
 			tr->langopts.unstressed_wd1 = 2;
-			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_HUNDRED_AND | NUM_DFRACTION_4;
+			tr->langopts.numbers = NUM_DECIMAL_COMMA | NUM_HUNDRED_AND_DIGIT | NUM_DFRACTION_4 | NUM_ZERO_HUNDRED;
 
 		}
 		break;
 
+	case L('w','o'):
+			tr->langopts.stress_rule = STRESSPOSN_1L;
+			tr->langopts.numbers = NUM_AND_UNITS | NUM_HUNDRED_AND | NUM_OMIT_1_HUNDRED | NUM_OMIT_1_THOUSAND | NUM_SINGLE_STRESS;
+		break;
+
 	case L('z','h'):
 	case L_zhy:
 		{
@@ -1127,38 +1317,16 @@ SetLengthMods(tr,3);  // all equal
 
 
 
+
 static void Translator_Russian(Translator *tr)
 {//===========================================
 	static const unsigned char stress_amps_ru[] = {16,16, 18,18, 20,24, 24,22 };
 	static const short stress_lengths_ru[8] = {150,140, 220,220, 0,0, 260,280};
-
-
-	// character codes offset by 0x420
-	static const char ru_vowels[] = {0x10,0x15,0x31,0x18,0x1e,0x23,0x2b,0x2d,0x2e,0x2f,0};
-	static const char ru_consonants[] = {0x11,0x12,0x13,0x14,0x16,0x17,0x19,0x1a,0x1b,0x1c,0x1d,0x1f,0x20,0x21,0x22,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2c,0};
-	static const char ru_soft[] = {0x2c,0x19,0x27,0x29,0};   // letter group B  [k ts; s;]
-	static const char ru_hard[] = {0x2a,0x16,0x26,0x28,0};   // letter group H  [S Z ts]
-	static const char ru_nothard[] = {0x11,0x12,0x13,0x14,0x17,0x19,0x1a,0x1b,0x1c,0x1d,0x1f,0x20,0x21,0x22,0x24,0x25,0x27,0x29,0x2c,0};
-	static const char ru_voiced[] = {0x11,0x12,0x13,0x14,0x16,0x17,0};    // letter group G  (voiced obstruents)
-	static const char ru_ivowels[] = {0x2c,0x15,0x31,0x18,0x2e,0x2f,0};   // letter group Y  (iotated vowels & soft-sign)
+	static const char ru_ivowels2[] = {0x2c,0x15,0x18,0x2e,0x2f,0};   // add more vowels to letter group Y  (iotated vowels & soft-sign)
 
 	SetupTranslator(tr,stress_lengths_ru,stress_amps_ru);
-
-	tr->charset_a0 = charsets[18];   // KOI8-R
-	tr->transpose_offset = 0x42f;  // convert cyrillic from unicode into range 0x01 to 0x22
-	tr->transpose_min = 0x430;
-	tr->transpose_max = 0x451;
-
-	tr->letter_bits_offset = OFFSET_CYRILLIC;
-	memset(tr->letter_bits,0,sizeof(tr->letter_bits));
-	SetLetterBits(tr,0,ru_vowels);
-	SetLetterBits(tr,1,ru_soft);
-	SetLetterBits(tr,2,ru_consonants);
-	SetLetterBits(tr,3,ru_hard);
-	SetLetterBits(tr,4,ru_nothard);
-	SetLetterBits(tr,5,ru_voiced);
-	SetLetterBits(tr,6,ru_ivowels);
-	SetLetterBits(tr,7,ru_vowels);
+	SetCyrillicLetters(tr);
+	SetLetterBits(tr,6,ru_ivowels2);
 
 	tr->langopts.param[LOPT_UNPRONOUNCABLE] = 0x432;    // [v]  don't count this character at start of word
 	tr->langopts.param[LOPT_REGRESSIVE_VOICING] = 1;
